//+++2005-03-01
//    Copyright (C) 2004,2005  Mike Rieker, Beverly, MA USA
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; version 2 of the License.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//---2005-03-01

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;

public class HelpPane extends JPanel {
  private static final String helpindexfilename = "helpindex.txt";
  private static final String helptextfilenameprefix = "helppage_";
  private static final String helptextfilenamesuffix = ".txt";

  private JScrollPane helppagescrollpane;
  private JTextArea helppagetextarea;

  public static void main (String[] argv) {
    HelpPane hp = new HelpPane ();
    JFrame hf = new JFrame ("Help");
    hf.setDefaultCloseOperation (JFrame.EXIT_ON_CLOSE);
    hf.setContentPane (hp);
    hf.pack ();
    hf.setVisible (true);
  }

  public HelpPane () {
    setLayout (new BorderLayout ());
    add (new IndexPanel (), BorderLayout.WEST);
    helppagetextarea = new JTextArea (30, 50);
    helppagescrollpane = new JScrollPane (helppagetextarea);
    add (helppagescrollpane, BorderLayout.CENTER);
  }

  /* This panel is used to look at the help index */

  private class IndexPanel extends JPanel implements CaretListener {
    private Ikwn[] indexkeywordnumberarray;    // one of these per all keyword entries
    private Ikws[] indexkeywordstringarray;    // one of these per major keyword entry
    private int highlightedkeyword = -1;       // index in indexkeywordnumberarray that's highlighed
    private JPanel indexkeywordtextarea;
    private JTextField indexkeywordtextfield;
    private String lastvalue = "";
    private VerticalScrollPane indexkeywordscrollpane;

    public IndexPanel () {
      ArrayList ikwna = new ArrayList ();
      BufferedReader keywordfile;
      Font textfieldfont;
      int i, j, kwi;
      int width = 0;
      String keywordbare, keywordline;
      TreeSet ikwsa = new TreeSet ();

      textfieldfont = new JTextField (10).getFont ();

      /* Create arrays and scroll pane from index file */

      try {

        /* Open the index file and create a text area to put the index's keywords in */

        keywordfile = new BufferedReader (new InputStreamReader (new FileInputStream (helpindexfilename)));
        indexkeywordtextarea = new JPanel (new GridLayout (0, 1));
        indexkeywordtextarea.setBackground (Color.WHITE);

        /* Read index file line by line */

        kwi = 0;
        while ((keywordline = keywordfile.readLine ()) != null) {

          /* The tab near the end separates the keyword from the filenumber#linenumber */

          for (i = keywordline.length (); -- i >= 0;) {
            if (keywordline.charAt (i) <= ' ') break;
          }
          if (i > 0) {
            for (j = i; -- j > 0;) {
              if (keywordline.charAt (j-1) > ' ') break;
            }
            keywordbare = keywordline.substring (0, j);

            /* Get the split between filenumber and linenumber */

            j = keywordline.substring (i + 1).indexOf ('#');
            if (j > 0) {
              if (width < i) width = i;

              /* Add keyword to scroll area and add filenumber and linenumber to the index keyword number array */

              Ikwn ikwn = new Ikwn (kwi, textfieldfont, keywordbare,
                                    keywordline.substring (i + 1, i + 1 + j),
                                    keywordline.substring (i + 1 + j + 1));

              indexkeywordtextarea.add (ikwn);
              ikwna.add (ikwn);

              /* If major keyword (ie, line doesn't begin with a space), add it to searchable index */

              if (keywordline.charAt (0) > ' ') {
                ikwsa.add (new Ikws (keywordbare.toLowerCase (), kwi));
              }
              kwi ++;
            }
          }
        }

        /* Done reading, close file */

        keywordfile.close ();

      } catch (IOException ioe) {
        JOptionPane.showMessageDialog (this, "Error reading " + helpindexfilename + ": " + ioe.getMessage ());
        return;
      }

      /* Convert the lists into fixed-sized arrays for ease of use */

      i = ikwna.size ();
      indexkeywordnumberarray = new Ikwn[i];
      while (-- i >= 0) indexkeywordnumberarray[i] = (Ikwn) ikwna.get (i);

      indexkeywordstringarray = new Ikws[ikwsa.size()];
      Iterator it = ikwsa.iterator ();
      i = 0;
      while (it.hasNext ()) indexkeywordstringarray[i++] = (Ikws) it.next ();

      /* Create the index panel */

      setLayout (new BorderLayout ());

      JPanel upperpanel = new JPanel (new GridLayout (0, 1));

      upperpanel.add (new JLabel ("Enter keyword:"));

      indexkeywordtextfield = new JTextField (width);
      indexkeywordtextfield.addCaretListener (this);
      upperpanel.add (indexkeywordtextfield);

      upperpanel.add (new JLabel ("Click on desired entry:"));

      add (upperpanel, BorderLayout.NORTH);

      indexkeywordscrollpane = new VerticalScrollPane (indexkeywordtextarea, indexkeywordnumberarray.length);
      add (indexkeywordscrollpane, BorderLayout.CENTER);
    }

    /* This routine is called when there is a change to the index keyword text field */

    public void caretUpdate (CaretEvent ce) {
      int j;
      String newvalue;

      /* See if there is a change in the input string */

      newvalue = indexkeywordtextfield.getText ().toLowerCase ().trim ();
      if (!newvalue.equals (lastvalue)) {
        lastvalue = newvalue;

        /* Look for entry in major keyword list that is just barely .ge. input value */

        for (j = 0; j < indexkeywordstringarray.length; j ++) {
          if (indexkeywordstringarray[j].keyword.compareTo (lastvalue) >= 0) break;
        }

        /* Set the scroll area index = keyword found and highlight it in scroll area */

        if (j < indexkeywordstringarray.length) {
          selectline (indexkeywordstringarray[j].index);
        }
      }
    }

    /* A particular line in the keyword scrollable area was selected, highlight it */

    private void selectline (int i) {
      indexkeywordscrollpane.setvalue (i);
      if (highlightedkeyword >= 0) {
        indexkeywordnumberarray[highlightedkeyword].selected (false);
      }
      highlightedkeyword = i;
      indexkeywordnumberarray[highlightedkeyword].selected (true);
    }

    /* Display a partiular file scrolled to a particular line */

    private String currentfilenumber;
    private HashMap currentfilehashmap;

    private void displayfile (String filenumber, String linenumber) {

      /* If we're not currently displaying the correct file, read it and display it */

      if ((currentfilenumber == null) || !currentfilenumber.equals (filenumber)) {
        String filename = helptextfilenameprefix + filenumber + helptextfilenamesuffix;
        try {

          /* Open the new help file */

          BufferedReader helpfile = new BufferedReader (new InputStreamReader (new FileInputStream (filename)));

          /* Clear out everything to prepare for reading */

          currentfilehashmap = new HashMap ();
          helppagetextarea.setText ("");
          int idx = 0;
          String line;

          /* Process lines from file one at a time */

          while ((line = helpfile.readLine ()) != null) {

            /* Remove any #linenumber# tags from beginning of line and add to hash table */

            while ((line.length () > 0) && (line.charAt (0) == '#')) {
              int len = line.substring (1).indexOf ('#');
              if (len < 0) len = line.length () - 1;
              if (len > 0) {
                currentfilehashmap.put (line.substring (1, len + 1), new Integer (idx));
              }
              line = line.substring (len + 1);
            }

            /* Put line on end of text area */

            helppagetextarea.append (line);
            helppagetextarea.append ("\n");
            idx ++;
          }

          /* Done, close file and remember which file we have on display now */

          helpfile.close ();
          currentfilenumber = filenumber;
        } catch (IOException ioe) {

          /* Read error, remember we don't have a file and display error message */

          currentfilenumber  = null;
          currentfilehashmap = null;
          helppagetextarea.setText ("");
          JOptionPane.showMessageDialog (this, "Error reading " + filename + ": " + ioe.getMessage ());
          return;
        }
      }

      /* Find which line the #linenumber# tag is on and move to it */

      Object lineindex = currentfilehashmap.get (linenumber);
      if (lineindex != null) {
        helppagescrollpane.getVerticalScrollBar ().setValue (((Integer)lineindex).intValue ());
      }
    }

    /* There is one of these for each keyword read from the index file */

    private class Ikwn extends JLabel implements MouseListener {
      public String filenumber; // which help file the contents are in
      public String linenumber; // which line of the file to display at the top
      public int index;         // index of this element in indexkeywordnumberarray

      /* Constructor creates a label with white background and black foreground      */
      /* Use same font as the keyword entry text field box and set up mouse listener */

      public Ikwn (int i, Font f, String kw, String fn, String ln) {
        super (kw);
        index = i;
        filenumber = fn;
        linenumber = ln;
        setFont (f);
        setForeground (Color.BLACK);
        setBackground (Color.WHITE);
        addMouseListener (this);
      }

      /* Highlight if selected */

      public void selected (boolean s) {
        setForeground (s ? Color.RED : Color.BLACK);
      }

      /* If user clicks on the keyword, select it */

      public void mouseClicked (MouseEvent me) {
        if (me.getButton () == MouseEvent.BUTTON1) {   // only respond to mouse button 1
          indexkeywordtextfield.setText (getText ());  // this causes caretUpdate to be called
          displayfile (filenumber, linenumber);        // display particular file and line
        }
      }

      /* Dummy these methods for the MouseListener */

      public void mouseEntered (MouseEvent me) {}
      public void mouseExited (MouseEvent me) {}
      public void mousePressed (MouseEvent me) {}
      public void mouseReleased (MouseEvent me) {}
    }

    /* There is one of these for each major keyword read from the index file */

    public class Ikws implements Comparable {
      public String keyword;  // keyword string (lower case)
      public int index;       // index in ikwna = scroll area line number (starting at 0)

      public Ikws (String kw, int ix) {
        keyword = kw;
        index = ix;
      }

      public int compareTo (Object o) {
        return (keyword.compareTo (((Ikws)o).keyword));
      }
    }
  }

  private class SearchKeywordTextField extends JTextField implements CaretListener {
    private String lastvalue = "";
    public void caretUpdate (CaretEvent ce) {
      String newvalue = getText ().toLowerCase ();
      if (!newvalue.equals (lastvalue)) {
        lastvalue = newvalue;
      }
    }
  }

  /* Create a scroll pane with only a vertical scrollbar */
  /* Make the scrollbar settable                         */

  private class VerticalScrollPane extends JScrollPane {
    private JScrollBar vscrollbar;
    public VerticalScrollPane (JComponent view, int ntotal) {
      super (view, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
      vscrollbar = getVerticalScrollBar ();
      vscrollbar.setMaximum (ntotal - 1);
      vscrollbar.setMinimum (0);
    }
    public void setvalue (int value) {
      vscrollbar.setValue (value);
    }
  }
}
