//+++2004-11-01
//    Copyright (C) 2004  Mike Rieker, Beverly, MA USA
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; version 2 of the License.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//---2004-11-01

// Performs multiplies and divides

public class IODevMuldiv extends IODev {

  private int div, dh, dl, m1, m2;

  public IODevMuldiv () {
    super (-32, 0xFFFFF020); // register space is 32 bytes starting at 0xFFFFF020

    SparcAsm.defpermasym ("IODevMulM1",  super.base +  0); // WO multiplier
    SparcAsm.defpermasym ("IODevMulM2",  super.base +  4); // WO multiplier
    SparcAsm.defpermasym ("IODevMulUPH", super.base +  0); // RO unsigned product hi-order
    SparcAsm.defpermasym ("IODevMulUPL", super.base +  4); // RO unsigned product lo-order
    SparcAsm.defpermasym ("IODevMulSPH", super.base +  8); // RO signed product hi-order
    SparcAsm.defpermasym ("IODevMulSPL", super.base + 12); // RO signed product lo-order

    SparcAsm.defpermasym ("IODevDivDH",  super.base + 16); // WO dividend hi-order
    SparcAsm.defpermasym ("IODevDivDL",  super.base + 20); // WO dividend lo-order
    SparcAsm.defpermasym ("IODevDivDiv", super.base + 24); // WO divisor
    SparcAsm.defpermasym ("IODevDivUQ",  super.base + 16); // RO unsigned quotient
    SparcAsm.defpermasym ("IODevDivUR",  super.base + 20); // RO unsigned remainder
    SparcAsm.defpermasym ("IODevDivSQ",  super.base + 24); // RO signed quotient
    SparcAsm.defpermasym ("IODevDivSR",  super.base + 28); // RO signed remainder
  }

  public int readreg (int reg, int size) {
    if (size != 4) return (0);
    switch (reg) {

      // Multiplication

      case  0: { // unsigned product hi-order
        long x = m1 & 0xFFFFFFFFL;
        long y = m2 & 0xFFFFFFFFL;
        long z = x * y;
        return ((int) (z >> 32));
      }
      case  4: { // unsigned product lo-order
        long x = m1 & 0xFFFFFFFFL;
        long y = m2 & 0xFFFFFFFFL;
        long z = x * y;
        return ((int) z);
      }
      case  8: { // signed product hi-order
        long x = m1;
        long y = m2;
        long z = x * y;
        return ((int) (z >> 32));
      }
      case 12: { // signed product lo-order
        long x = m1;
        long y = m2;
        long z = x * y;
        return ((int) z);
      }

      // Division

      case 16: { // unsigned quotient
        long d = (((long) dh) << 32) + (((long) dl) & 0xFFFFFFFFL);
        long v = ((long) div) & 0xFFFFFFFFL;
        if (v == 0) return (0);
        if (d >= 0) return ((int) (d / v));
        for (int i = 64; -- i >= 0;) {
          if (((d >> 32) & 0xFFFFFFFFL) < v) d += d;
          else d += d + 1 - (v << 32);
        }
        return ((int) d);
      }
      case 20: { // unsigned remainder
        long d = (((long) dh) << 32) + (((long) dl) & 0xFFFFFFFFL);
        long v = ((long) div) & 0xFFFFFFFFL;
        if (v == 0) return (0);
        if (d >= 0) return ((int) (d % v));
        for (int i = 64; -- i >= 0;) {
          if (((d >> 32) & 0xFFFFFFFFL) < v) d += d;
          else d += d + 1 - (v << 32);
        }
        return ((int) (d >> 32));
      }
      case 24: { // signed quotient
        long d = (((long) dh) << 32) + (((long) dl) & 0xFFFFFFFFL);
        if (div == 0) return (0);
        return ((int)(d / div));
      }
      case 28: { // signed remainder
        long d = (((long) dh) << 32) + (((long) dl) & 0xFFFFFFFFL);
        if (div == 0) return (0);
        return ((int)(d % div));
      }

      // Who knows?

      default: return (0);
    }
  }

  public void writereg (int reg, int size, int data) {
    if (size == 4) switch (reg) {
      case  0: { // multipliers
        m1 = data;
        break;
      }
      case  4: {
        m2 = data;
        break;
      }
      case 16: { // dividend hi-order
        dh = data;
        break;
      }
      case 20: { // dividend lo-order
        dl = data;
        break;
      }
      case 24: { // divisor
        div = data;
        break;
      }
    }
  }
}
