//+++2005-02-28
//    Copyright (C) 2004,2005  Mike Rieker, Beverly, MA USA
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; version 2 of the License.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//---2005-02-28

/********************************************************************/
/* Override the JDesktopPane's paint to put up a special background */
/********************************************************************/

import java.awt.*;
import java.awt.image.*;
import java.net.*;
import javax.swing.*;

public class MainPane extends JDesktopPane implements ImageObserver {

  private Image image = null;
  private int image_height = 0;
  private int image_width  = 0;
  private String picfilename;
  private static final double picscales[] = { 0.25, 0.50, 0.75, 1.00, 1.50, 2.00 };

  /* Constructor starts reading the image in     */
  /* Input: pfn = image filename within jar file */

  public MainPane (String pfn) {
    String jarfilename = System.getProperty ("java.class.path");                              // eg, blabla/SparcSim.jar;someothercrap
    String filesepchar = System.getProperty ("file.separator");                               // eg, \ or /
    String pathsepchar = System.getProperty ("path.separator");                               // eg, ;

    picfilename = pfn;
    if ((jarfilename != null) && (filesepchar != null) && (pathsepchar != null)) {
      int pathsepindx = jarfilename.indexOf (pathsepchar.charAt (0));                         // find the ; before ;someothercrap
      if (pathsepindx > 0) jarfilename = jarfilename.substring (0, pathsepindx);              // chop at first ;
      int jfnl = jarfilename.length ();
      if ((jfnl > 4) && jarfilename.substring (jfnl - 4).equals (".jar")) {                   // see if it's a jar file (instead of directory with class files)
        String imageurlstring = "jar:file:" + jarfilename + "!/" + picfilename;               // jar file, create a jar URL string (see JarURLConnection)
        try {
          URL imageurl = new URL (imageurlstring);                                            // create URL object
          image = Toolkit.getDefaultToolkit ().getImage (imageurl);                           // start opening the image using that URL
        } catch (MalformedURLException mue) {
          System.err.println ("Malformed URL " + imageurlstring + ": " + mue.getMessage ());  // oh well!
        }
      } else {
        String imagefilename = jarfilename + filesepchar + picfilename;                       // not a jar, assume it's a directory
        image = Toolkit.getDefaultToolkit ().getImage (imagefilename);                        // start opening image using <directory><filesep><picfilename>
      }
    }
  }

  /* This is from the 'ImageObserver' - it gets called as more of the image gets read in */

  public boolean imageUpdate (Image img, int infoflags, int x, int y, int width, int height) {
    if ((infoflags & ImageObserver.ERROR) != 0) {        // see if error reading image file
      image = null;                                      // if so, tell paint routine not to bother
      System.err.println ("Error reading background image");
      return (false);                                    // ... and don't try to read more of image file
    }
    if ((infoflags & ImageObserver.HEIGHT) != 0) image_height = height;
    if ((infoflags & ImageObserver.WIDTH)  != 0) image_width  = width;
    repaint ();                                          // repaint screen with what we have now
    return ((infoflags & ImageObserver.ALLBITS) == 0);   // maybe we need more of the image
  }

  /* This gets called when the GUI system needs to repaint the screen */

  public void paint (Graphics g) {

    /* Fill plain background */

    int panel_height = getHeight ();                                   // get panel dimensions
    int panel_width  = getWidth ();
    g.setColor (getBackground ());                                     // fill in background
    g.fillRect (0, 0, panel_width, panel_height);

    /* Draw picture background */

    if (image != null) {                                               // see if image is present (and not errored)
      if ((image_height > 0) && (image_width > 0)) {
        for (int i = picscales.length; -- i >= 0;) {
          int scaled_height = (int) (picscales[i] * image_height);
          int scaled_width  = (int) (picscales[i] * image_width);
          if ((scaled_height <= panel_height) && (scaled_width <= panel_width)) {
            int x_offset = (panel_width  - scaled_width)  / 2;
            int y_offset = (panel_height - scaled_height) / 2;
            g.drawImage (image, x_offset, y_offset, scaled_width,      // draw whatever we have of the image so far
                         scaled_height, getBackground (), this);
            break;
          }
        }
      } else {
        g.drawImage (image, 0, 0, 1, 1, getBackground (), this);       // don't know the size yet, draw real small
      }
    }

    /* Draw sub-panels */

    super.paintChildren (g);
  }
}
