//+++2005-02-28
//    Copyright (C) 2004,2005  Mike Rieker, Beverly, MA USA
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; version 2 of the License.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//---2005-02-28

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;

// This class is the main class for the Simulator

public class SparcSim extends JFrame {

  ////////////////////
  // STATIC SECTION //
  ////////////////////

  public static final String namever = "SparcSim v38";
  public static final String aboutit =
        namever + " - Sparc Simulator\n" +
        "Copyright 2004,2005, Mike Rieker, Beverly, MA, USA\n" +
        "EXPECT it to FAIL when someone's HeALTh or PROpeRTy is at RISk\n" +
        "See http://www.o3one.org/sparcsim.html\n" +
        "\n" +
        "This program is free software; you can redistribute it and/or modify\n" +
        "it under the terms of the GNU General Public License as published by\n" +
        "the Free Software Foundation; version 2 of the License.\n" +
        "See http://www.gnu.org/licenses/gpl.html\n" +
        "\n" +
        "This program is distributed in the hope that it will be useful,\n" +
        "but WITHOUT ANY WARRANTY; without even the implied warranty of\n" +
        "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n" +
        "GNU General Public License for more details.";

  public static BigEndianMemory memory;       // this points to the one-and-only memory object
  public static Font monofont;                // lots of things use this for their monospaced font
  private static MainPane mainpane;           // put JInternalPanes in here

  // The sourceeditorlist, sparccpulist and memwindowlist are synchronized access only
  // ... because the 'continue' function of the CPUs executes as a separate
  //     threads and may try to access these lists while they are running

  private static LinkedList sourceeditorlist; // list of all source editor frames
  private static LinkedList sparccpulist;     // list of all Sparc CPU frames
  private static LinkedList memwindowlist;    // list of all memory window frames

  private static int[] breakpointlist;        // bitmap of armed breakpoints
                                              // although accessed by more than one thread,
                                              // it doesn't move, so it need not be synchronized

  private static int minwindowx, newwindowx;  // where to put new windows on pseudo-desktop
  private static int minwindowy, newwindowy;

  // This is the real 'main' program
  // The args are a list of source files to open in source editor windows on startup

  public static void main (String[] args) {
    sparccpulist = new LinkedList ();
    sourceeditorlist = new LinkedList ();
    memwindowlist = new LinkedList ();
    monofont = new Font ("Monospaced", Font.PLAIN, 12);

    File gpl = new File ("SparcSim.gpl");
    if (!gpl.exists ()) {
      if (JOptionPane.showConfirmDialog (null, aboutit,
          namever + " License", JOptionPane.OK_CANCEL_OPTION) != JOptionPane.OK_OPTION) System.exit(0);
      try {
        gpl.createNewFile();
      } catch (IOException ioe) {
      }
    }

    // Create main control panel window
    // There is only ever one of these

    new SparcSim ();

    // Create memory and a display window

    BigEndianMemory.setsize (1024*1024);
    BigEndianMemory bem = new BigEndianMemory ();
    bem.addInternalFrameListener (new FrameCloseListener (bem, memwindowlist));
    mainpane.add (bem);
    bem.setLocation (whereitsat ());
    synchronized (memwindowlist) {
      memwindowlist.addLast (bem);
    }
    breakpointlist = new int[BigEndianMemory.getsize()/128];

    // Create memory-mapped pseudo-IO devices

    new IODevPrint ();
    new IODevMuldiv ();

    // Create one CPU for starters

    SparcCPU sc = new SparcCPU ();
    sc.addInternalFrameListener (new FrameCloseListener (sc, sparccpulist));
    mainpane.add (sc);
    sc.setLocation (whereitsat ());
    synchronized (sparccpulist) {
      sparccpulist.addLast (sc);
    }

    // Create one source editor for each arg given
    // If no args, create a blank source editor

    if (args.length <= 0) {
      SourceEditor se = new SourceEditor ();
      se.addInternalFrameListener (new FrameCloseListener (se, sourceeditorlist));
      mainpane.add (se);
      se.setLocation (whereitsat ());
      synchronized (sourceeditorlist) {
        sourceeditorlist.addLast (se);
      }
    } else {
      for (int i = 0; i < args.length; i ++) {
        SourceEditor se = new SourceEditor ();
        se.addInternalFrameListener (new FrameCloseListener (se, sourceeditorlist));
        mainpane.add (se);
        se.setLocation (whereitsat ());
        se.openfile (args[i]);
        synchronized (sourceeditorlist) {
          sourceeditorlist.addLast (se);
        }
      }
    }
  }

  // Add an internal frame to the main pane

  public static void addtomainpane (JInternalFrame jif) {
    mainpane.add (jif);
    jif.setLocation (whereitsat ());
  }

  // Figure out where to create a new window within the pseudo desktop

  private static Point whereitsat () {

    // Make sure at least the upper left corner will
    // be within the current pseudo desktop

    if ((newwindowx + 30 > mainpane.getWidth ())
     || (newwindowy + 30 > mainpane.getHeight ())) {
      newwindowx = minwindowx;
      newwindowy = minwindowy;
    }

    // Return point and increment for next time

    Point p = new Point (newwindowx, newwindowy);
    newwindowx += 30;
    newwindowy += 30;
    return (p);
  }

  // This routine gets called when a 'compile' button
  // is clicked on any source editor window

  public static void compileall () {
    boolean pass2;
    int i;
    SourceEditor se;
    SparcCPU sc;

    // Compile all source files in case their sizes changed

    SparcAsm.reset ();
    pass2 = false;
    while (true) {
      SparcAsm.currentlocation = 0x4000;
      i = 0;
      while (true) {
        se = null;
        synchronized (sourceeditorlist) {
          if (i < sourceeditorlist.size ()) {
            se = (SourceEditor) sourceeditorlist.get (i);
          }
        }
        if (se == null) break;
        SparcAsm.assemble (se, pass2, SparcAsm.currentlocation);
        i ++;
      }
      if (pass2) break;
      pass2 = true;
    }

    // Clear all breakpoints as addresses and sourcelines are probably all different now

    clearallbreakpoints ();

    // Reset all CPUs - this also sets their PCs to the _start address

    i = 0;
    while (true) {
      sc = null;
      synchronized (sparccpulist) {
        if (i < sparccpulist.size ()) {
          sc = (SparcCPU) sparccpulist.get (i);
        }
      }
      if (sc == null) break;
      sc.reset ();
      i ++;
    }
  }

  // CPUs call this to check to see if a breakpoint has been hit
  // This routine must be threadsafe

  public static boolean checkbreakpoints (int address) {
    boolean hit = false;

    int i = address / 128;
    if (i < breakpointlist.length) {
      hit = ((breakpointlist[i] >> ((address / 4) & 31)) & 1) != 0;
    }
    return (hit);
  }

  public static void setbreakpoint (int address) {
    int i = address / 128;
    if (i < breakpointlist.length) breakpointlist[i] |= 1 << ((address / 4) & 31);
    else JOptionPane.showMessageDialog (null, "Attempting to set breakpoint at invalid address "
                                              + HexTextField.hexstring (address));
  }

  public static void clearbreakpoint (int address) {
    int i = address / 128;
    if (i < breakpointlist.length) breakpointlist[i] &= ~ (1 << ((address / 4) & 31));
  }

  public static void clearallbreakpoints () {
    if (breakpointlist != null) {
      synchronized (breakpointlist) {
        for (int i = breakpointlist.length; -- i >= 0;) breakpointlist[i] = 0;
      }
    }
  }

  // CPUs call this when they halt to update the source editors on where they are
  // This routine must be threadsafe

  public static void updateselocation () {
    int [] addresses;

    // Get list of addresses that all the CPUs are at

    synchronized (sparccpulist) {
      addresses = new int[sparccpulist.size()];
      for (int i = addresses.length; --i >= 0;) {
        SparcCPU sc = (SparcCPU) sparccpulist.get (i);
        addresses[i] = sc.getregister (SparcCPU.PC);
      }
    }

    // Update source editors to display the line in green

    synchronized (sourceeditorlist) {
      for (int i = sourceeditorlist.size (); --i >= 0; ) {
        SourceEditor se = (SourceEditor) sourceeditorlist.get (i);
        se.updatecpulocation (addresses);
      }
    }
  }

  // Update all memory displays of the given 32-bit aligned address

  public static void updatememwindows (int address) {
    int i = 0;
    while (true) {
      BigEndianMemory bem = null;
      synchronized (memwindowlist) {                      // make sure memwindowlist doesn't change on us
        if (i < memwindowlist.size ()) {                  // see if we're within limits
          bem = (BigEndianMemory) memwindowlist.get (i);  // if so, get a memory frame from the list
        }
      }
      if (bem == null) break;                             // all done if no more frames
      bem.updatedisplay (address);
      i ++;                                               // see if there are more frames to update
    }
  }

  //  This routine is called to exit the whole simulator
  //    Input:
  //      yes_no_option = YES_NO_CANCEL_OPTION : user is given option to cancel the exit
  //                      YES_NO_OPTION : user is not given option to cancel the exit
  //    Output:
  //      returns only if user clicked cancel button on one of the source file saves
  //      ... otherwise it exits

  private static void foad (int yes_no_option) {

    // Make sure there aren't any unsaved source changes the user wants

    synchronized (sourceeditorlist) {
      for (int i = 0; i < sourceeditorlist.size (); i ++) {
        SourceEditor se = (SourceEditor) sourceeditorlist.get (i);
        if (!se.maybesavechanges (yes_no_option)) return;
      }
    }

    // That's all we check for, breathe out, don't breathe in

    System.exit (0);
  }

  ////////////////////
  // OBJECT SECTION //
  ////////////////////

  // Constructor for the main frame (pseudo-desktop)
  // that encompasses all things drawn on the screen
  // There is only ever one of these

  public SparcSim () {
    super (namever);

    // Create the pseudo-desktop to almost fill the actual desktop

    Dimension screensize = Toolkit.getDefaultToolkit ().getScreenSize ();
    super.setBounds (30, 30, screensize.width - 60, screensize.height - 60);
    mainpane = new MainPane ("salem1952.jpg");
    super.setContentPane (mainpane);

    // Create the main control button panel - put it in the menubar area of the main window

    JMenuBar buttonpanel = new JMenuBar ();
    super.setJMenuBar (buttonpanel);
    buttonpanel.setLayout (new FlowLayout ());

    JButton sourceeditorbutton = new JButton ("Open Source Editor");
    sourceeditorbutton.addActionListener (new SourceEditorButtonListener ());
    JButton sparccpubutton = new JButton ("Open Sparc CPU");
    sparccpubutton.addActionListener (new SparcCPUButtonListener ());
    JButton memorybutton = new JButton ("Open Memory Viewer");
    memorybutton.addActionListener (new MemoryButtonListener ());
    JButton exitbutton = new JButton ("Exit Simulator");
    exitbutton.addActionListener (new ExitButtonListener ());
    JButton aboutbutton = new JButton ("About");
    aboutbutton.addActionListener (new AboutButtonListener ());

    buttonpanel.add (sourceeditorbutton);
    buttonpanel.add (sparccpubutton);
    buttonpanel.add (memorybutton);
    buttonpanel.add (exitbutton);
    buttonpanel.add (aboutbutton);

    minwindowx = newwindowx = 30;
    minwindowy = newwindowy = 0;

    // Add a listener so we know if the user is closing the whole pseudo-desktop

    this.addWindowListener (new MainWindowListener ());
    this.setVisible (true);
  }

  // This gets called when the Source Editor button is clicked
  // It opens a new source editor window and adds it to the list of source editors

  private class SourceEditorButtonListener implements ActionListener {
    public void actionPerformed (ActionEvent e) {
      SourceEditor se = new SourceEditor ();
      se.addInternalFrameListener (new FrameCloseListener (se, sourceeditorlist));
      mainpane.add (se);
      se.setLocation (whereitsat ());
      se.moveToFront ();
      synchronized (sourceeditorlist) {
        sourceeditorlist.addLast (se);
      }
    }
  }

  // This gets called when the Sparc CPU button is clicked
  // It opens a new Sparc CPU window and adds it to the list of Sparc CPUs

  private class SparcCPUButtonListener implements ActionListener {
    public void actionPerformed (ActionEvent e) {
      SparcCPU sc = new SparcCPU ();
      sc.addInternalFrameListener (new FrameCloseListener (sc, sparccpulist));
      mainpane.add (sc);
      sc.setLocation (whereitsat ());
      sc.moveToFront ();
      synchronized (sparccpulist) {
        sparccpulist.addLast (sc);
      }
    }
  }

  // This gets called when the Memory Window button is clicked
  // It opens a new Memory window and adds it to the list of memory windows

  private class MemoryButtonListener implements ActionListener {
    public void actionPerformed (ActionEvent e) {
      BigEndianMemory bem = new BigEndianMemory ();
      bem.addInternalFrameListener (new FrameCloseListener (bem, memwindowlist));
      mainpane.add (bem);
      bem.setLocation (whereitsat ());
      bem.moveToFront ();
      synchronized (memwindowlist) {
        memwindowlist.addLast (bem);
      }
    }
  }

  // This is called when the user clicks the 'Exit Simulator' button

  private class ExitButtonListener implements ActionListener {
    public void actionPerformed (ActionEvent e) {
      if (JOptionPane.showConfirmDialog (null, "Are you sure you want to exit?", "Exit Button",
                                         JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION) {
        foad (JOptionPane.YES_NO_CANCEL_OPTION);
      }
    }
  }

  // This is called when the user clicks the 'About' button

  private class AboutButtonListener implements ActionListener {
    public void actionPerformed (ActionEvent e) {
      JOptionPane.showMessageDialog (null, aboutit, "About " + namever,
                                     JOptionPane.INFORMATION_MESSAGE);
    }
  }

  // This routine is called when the main window is closed
  // It exits the whole schmeel
  // We don't prompt the user as there is no way to undo the close
  // ... but we can ask them if they want modified source files saved

  private class MainWindowListener extends WindowAdapter {
    public void windowClosing (WindowEvent e) {
      foad (JOptionPane.YES_NO_OPTION);
    }
  }

  // If someone closes a sub-window, remove it from its list

  private static class FrameCloseListener extends InternalFrameAdapter {
    JInternalFrame internalframe;
    LinkedList linkedlist;
    public FrameCloseListener (JInternalFrame jif, LinkedList ll) {
      internalframe = jif;
      linkedlist    = ll;
    }
    public void internalFrameClosing (InternalFrameEvent e) {
      synchronized (linkedlist) {
        linkedlist.remove (internalframe);
      }
    }
  }
}
