!
!  Perform Quicksort of 'int' array using the following algorithm:
!
!  qsortints (a, low, high)
!    if (low < high)
!      i = low
!      j = high
!      pivot = a[high]
!      repeat
!        while ((itemp = a[i]) < pivot) i ++
!        while (j > i) if ((jtemp = a[--j]) <= pivot) break
!        if (j <= i) break
!        a[i] = jtemp
!        a[j] = itemp
!      endrepeat
!      a[i] = pivot
!      a[high] = itemp
!      qsortints (a, low, i - 1)
!      qsortints (a, i + 1, high)
!    endif
!  return
!
	!!.text
	.global	qsortints
qsortints:
	cmp	%o1,%o2		! see if there's anything to sort
	bge	alldone		! if not, we're all done
	save	%sp,-96,%sp	! (meanwhile, make call frame)
startpass:
	sll	%i1,2,%l1	! point %l1 to a[low] (first element to sort)
	sll	%i2,2,%l2	! point %l2 to a[high] (last element, inclusive)
	add	%l1,%i0,%l1
	add	%l2,%i0,%l2
	mov	%l2,%l0		! save pointer to last element = pointer to pivot
	ld	[%l2],%o0	! get %o0 = pivot = a[high]
!
!  Scan forward from low end until we find value .ge. pivot
!  We don't have to put an upper limit test on the pointer because it will hit the pivot
!
scanltpivot:
	ld	[%l1],%o1	! get value from low end
	cmp	%o1,%o0		! compare to pivot
	bl,a	scanltpivot	! keep scanning until we have value .ge. pivot
	add	%l1,4,%l1	! (increment pointer only if we loop back)
				! %l1 = points to terminating value (.ge. pivot)
				! %o1 = contains terminating value
!
!  Scan backward from high end until we find value .le. pivot
!  We have to test lower limit on pointer in case all values are .gt. pivot
!
scangtpivot:
	cmp	%l2,%l1		! make sure we haven't collapsed yet
	bleu	endpass		! if so, we're done with this pass
	nop
	ld	[%l2-4],%o2	! get value from high end (predecrement)
	cmp	%o2,%o0		! compare to pivot
	bg	scangtpivot	! keep scanning until we have value .le. pivot
	sub	%l2,4,%l2	! (always decrement pointer)
				! %l2 = points to terminating value (.le. pivot)
				! %o2 = contains terminating value
!
!  %l1 points to value near the beginning that is .ge. pivot
!  %l2 points to value near the end that is .le. pivot
!  Swap values so all values near beginning are .le. pivot and all values near end are .ge. pivot
!
	st	%o1,[%l2]	! swap the two terminating values
	b	scanltpivot	! ... and resume scanning
	st	%o2,[%l1]
!
!  All values before %l1 are .le. pivot; all values after %l1 are .ge. pivot
!  Value at %l1 is .ge. pivot so swap it with pivot in case it is .gt. pivot
!
endpass:
	st	%o0,[%l1]	! store pivot in the middle
	st	%o1,[%l0]	! store high value on the end where pivot came from
!
!  Sort everything before the pivot then sort everything after the pivot
!
	subcc	%l1,%i0,%l1	! compute terminating 'i' index
	srl	%l1,2,%l1

	mov	%i0,%o0		! qsortints (a, low, i-1)
	mov	%i1,%o1
	call	qsortints
	sub	%l1,1,%o2

	add	%l1,1,%i1	! if (++ i < high) qsortints (a, i, high)
	cmp	%i1,%i2
	bl	startpass
	nop
alldone:
	ret
	restore
